/*
  alarm.h - Alarm Library for Arduino

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#pragma once

#include <stddef.h>
#include <stdint.h>
#include <stdbool.h>
#include "hardware.h"

class Alarm : protected Hardware
{
  public:

    typedef bool (*handler_t)(Alarm& alarm);

    Alarm& interval(int32_t interval, bool fromnow = true);
    int32_t interval(void);
    Alarm& handler(handler_t handler);
    handler_t handler(void);
    Alarm& param(uint8_t num, void *val);
    void *param(uint8_t num);
    bool expire(void);

  protected:

    friend class AlarmClass;

    Alarm    *_next;
    uint32_t  _alarm;
    int32_t   _interval;
    handler_t _handler;
    void     *_params[4];
};

class AlarmClass : protected Hardware
{
  public:
    AlarmClass(void) : _alarm_list(nullptr) {}
    void begin(void);
    bool start(Alarm& alarm);
    void ISR_ATTR cancel(Alarm& alarm);
    void ISR_ATTR cancel(uint8_t num, void *arg);
    void handle(void);

  protected:

    bool ISR_ATTR add(Alarm& alarm);

  private:

    bool ISR_ATTR timeup(void);

    friend void AlarmClass_timeup(void);

    Alarm *_alarm_list;
    bool _poll_mode;
};

extern AlarmClass AlarmTimer;
