/*
  config.h - Configuration

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#pragma once

#if defined(ARDUINO_ARCH_AVR) || defined(__AVR__)
  #define MCU_NAME avr
  #define DEFAULT_STACK_SIZE 128
  #define ISR_ATTR
#elif defined(__SAMD21__)
  #define MCU_NAME atsamd21
  #define DEFAULT_STACK_SIZE 1024
  #define ISR_ATTR
#elif defined(ARDUINO_ARCH_NRF52840) || defined(ARDUINO_SEEED_XIAO_NRF52840) || defined(ARDUINO_SEEED_XIAO_NRF52840_SENSE)
  #define MCU_NAME nrf52840
  #define __NRF52840__
  #define DEFAULT_STACK_SIZE 2048
  #define ISR_ATTR
#elif defined(ARDUINO_ARCH_RP2040) || defined(ARDUINO_RASPBERRY_PI_PICO) || defined(ARDUINO_ARCH_MBED_RP2040)
  #define MCU_NAME rp2040
  #define __RP2040__
  #define DEFAULT_STACK_SIZE 2048
  #define ISR_ATTR
#elif defined(ARDUINO_ARCH_ESP32) || defined(ESP32)
  #define MCU_NAME esp32
  #define __ESP32__
  #define DEFAULT_STACK_SIZE 1024
  #define ISR_ATTR IRAM_ATTR
#elif defined(ARDUINO_ARCH_ESP8266) || defined(ESP8266)
  #define MCU_NAME esp8266
  #define __ESP8266__
  #define DEFAULT_STACK_SIZE 1024
  #define ISR_ATTR IRAM_ATTR
#elif defined(ARDUINO_ARCH_STM32)
  #define MCU_NAME stm32
  #define __STM32__
  #define DEFAULT_STACK_SIZE 1024
  #define ISR_ATTR
#else
  #define MCU_NAME none
  #define __NONE__
  #define DEFAULT_STACK_SIZE 2048
  #define ISR_ATTR
#endif

#define _EXPAND1_(a)     a
#define _EXPAND2_(a,b)   _EXPAND1_(a)b
#define _EXPAND3_(a,b,c) _EXPAND2_(a,b)c
#define _STRINGIFY_(str) #str
#define _EXPANDSTR_(str) _STRINGIFY_(str)

