/*
  hardware.h - Hardware Library

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#pragma once

#include <stddef.h>
#include <stdint.h>
#include <stdbool.h>
#include "config.h"

#ifdef ARDUINO
  #include <Arduino.h>
#else
  #define INPUT 0
  #define OUTPUT 1
  #define INPUT_PULLUP 2
  #define HIGH 1
  #define LOW  0
#endif

#if ARDUINO_API_VERSION < 10300
  typedef uint8_t PinMode;
  typedef uint8_t PinStatus;
  #ifdef EXTENDED_PIN_MODE
  // Platforms who want to declare more than 256 pins need to define EXTENDED_PIN_MODE globally
    typedef uint32_t pin_size_t;
  #else
    typedef uint8_t pin_size_t;
  #endif
#endif

class Hardware
{
  public:

#ifdef ARDUINO
    static uint32_t micros(void) { return ::micros(); }
    static void pinMode(pin_size_t pin, PinMode mode) { ::pinMode(pin, mode); }
    static void digitalWrite(pin_size_t pin, PinStatus val) { ::digitalWrite(pin, val); }
    static PinStatus digitalRead(pin_size_t pin) { return ::digitalRead(pin); }
#else
    static uint32_t micros(void) { return 0; }
    static void pinMode(pin_size_t pin, PinMode mode) {}
    static void digitalWrite(pin_size_t pin, PinStatus val) {}
    static PinStatus digitalRead(pin_size_t pin) { return 0; }
#endif
    static size_t ISR_ATTR disableAndSaveInterrupts(void);
    static void ISR_ATTR restoreInterrupts(size_t save);
    static void idle(void);

  protected:

    /* For Alarm Class */
    static bool setupAlarmTimer(void);
    static void ISR_ATTR startAlarmTimer(int32_t interval);
};
