/*
  led.cpp - LED Library for Arduino

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include "led.h"

Led::CTRL Led::_leds[];

#define LED_NONE (pin_size_t)-1

void Led::begin(void)
{
  CTRL *ctrl = _leds;
  for (uint8_t i = 0; i < LED_COUNT; ++i, ++ctrl)
  {
    ctrl->pin    = LED_NONE;
    ctrl->off    = LOW;
    ctrl->mode   = MODE_DISABLE;
    ctrl->expire = false;
  }
}

void Led::setup(uint8_t led, pin_size_t pin, PinStatus off)
{
  CTRL& ctrl = _leds[led & LED_MASK];
  ctrl.pin    = pin;
  ctrl.off    = off;
  ctrl.mode   = MODE_DISABLE;
  ctrl.expire = false;
  digitalWrite(pin, off);
  pinMode(pin, OUTPUT);
  digitalWrite(pin, off);
}

void Led::off(uint8_t led, uint16_t ms)
{
  control(_leds[led & LED_MASK], MODE_OFF, ms);
}

void Led::on(uint8_t led, uint16_t ms)
{
  control(_leds[led & LED_MASK], MODE_ON, ms);
}

void Led::toggle(uint8_t led, uint16_t ms)
{
  control(_leds[led & LED_MASK], MODE_TOGGLE, ms);
}

void Led::control(CTRL& ctrl, MODE mode, uint16_t ms)
{
  ctrl.expire = false;
  if (ms)
  {
    ctrl.mode = mode;
    ctrl.alarm.interval(ms * 1000UL);
  }
  else
  {
    ctrl.mode = MODE_DISABLE;
    if (ctrl.pin != LED_NONE)
    {
      switch (mode)
      {
        case MODE_OFF:    digitalWrite(ctrl.pin, ctrl.off); break;
        case MODE_ON:     digitalWrite(ctrl.pin, ctrl.off ? LOW : HIGH); break;
        case MODE_TOGGLE: digitalWrite(ctrl.pin, digitalRead(ctrl.pin) ? LOW : HIGH); break;
        default: break;
      }
    }
  }
}

bool Led::status(uint8_t led)
{
  CTRL& ctrl = _leds[led & LED_MASK];
  return ctrl.pin != LED_NONE ? digitalRead(ctrl.pin) != ctrl.off : false;
}

bool Led::expire(uint8_t led)
{
  CTRL& ctrl = _leds[led & LED_MASK];
  bool rv = ctrl.expire;
  ctrl.expire = false;
  return rv;
}

void Led::handle(void)
{
  CTRL *ctrl = _leds;
  for (uint8_t i = 0; i < LED_COUNT; ++i, ++ctrl)
  {
    MODE mode = ctrl->mode;
    if ((mode != MODE_DISABLE) && ctrl->alarm.expire())
    {
      control(*ctrl, mode, 0);
      ctrl->expire = true;
      if (mode == MODE_TOGGLE)
        ctrl->mode = mode;
    }
  }
}
