/*
  led.h - LED Library for Arduino

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#pragma once

#include <stdint.h>
#include <stdbool.h>
#include "hardware.h"
#include "alarm.h"

#define LED_R 0 // Red
#define LED_G 1 // Green
#define LED_Y 2 // Yellow
#define LED_B 3 // Blue

#ifndef LED_BITS
#define LED_BITS  2
#endif
#define LED_COUNT (1 << LED_BITS)
#define LED_MASK  (LED_COUNT - 1)

class Led : protected Hardware
{
  public:

    static void begin(void);
    static void setup(uint8_t led, pin_size_t pin, PinStatus off);
    static void off(uint8_t led = 0, uint16_t ms = 0);
    static void on(uint8_t led_t = 0, uint16_t ms = 0);
    static void toggle(uint8_t led = 0, uint16_t ms = 0);
    static bool status(uint8_t led = 0);
    static bool expire(uint8_t led = 0);
    static void handle(void);

  private:

    typedef enum : uint8_t
    {
      MODE_DISABLE,
      MODE_OFF,
      MODE_ON,
      MODE_TOGGLE,
    } MODE;

    typedef struct
    {
      pin_size_t pin;
      PinStatus  off;
      MODE       mode;
      bool       expire;
      Alarm      alarm;
    } CTRL;

    static CTRL _leds[LED_COUNT];

    static void control(CTRL& ctrl, MODE mode, uint16_t ms);
};
