/*
  hardware.cpp - Hardware Library for ATSAMD21

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/

#include "../../config.h"

#ifdef __SAMD21__

#include "tcounter.h"
#include "../../hardware.h"
#include "../../alarm.h"

#define PRESCALER TCOUNTER_PRESCALER_DIV16  // 1/16 Clock
#define TIMER_US  (F_CPU / 16 / 1000000)

static TCounter *tc;
static int32_t tc_top;

inline void AlarmClass_timeup(void)
{
  AlarmTimer.timeup();
}

static void TimerCallback([[maybe_unused]] TCounter *obj, [[maybe_unused]] TCOUNTER_INT flag)
{
  AlarmClass_timeup();
}

bool Hardware::setupAlarmTimer(void)
{
  tc = TCFactory::alloc(TCOUNTER_MODE24, TCOUNTER_WAVEGEN_NFRQ);
  tc_top = tc->top();
  tc->setCallback(TimerCallback);
  tc->interrupt(TCOUNTER_INT_MC0, true);
  tc->prescaler(PRESCALER);
  tc->cc(TCOUNTER_MC0, 0);
  tc->count(0);
  tc->start();
  return true;
}

void Hardware::startAlarmTimer(int32_t interval)
{
  if (interval < 2)
    interval = TIMER_US;
  else if (interval < tc_top / TIMER_US)
    interval *= TIMER_US;
  else
    interval = tc_top;
  tc->cc(TCOUNTER_MC0, interval + tc->count());
}

size_t Hardware::disableAndSaveInterrupts(void)
{
  size_t save = __get_PRIMASK();
  __disable_irq();
  return save;
}

void Hardware::restoreInterrupts(size_t save)
{
  __set_PRIMASK(save);
}

void Hardware::idle(void)
{
  __enable_irq();
  __WFI();
}

#endif
