/*
  hardware.cpp - Hardware Library for RP2040

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include "../../config.h"

#ifdef __ESP32__

#include "esp32-hal.h"
#include "../../hardware.h"
#include "../../alarm.h"

#ifndef ALARM_TIMER_ID
#define ALARM_TIMER_ID 2
#endif

hw_timer_t *timer;

void ISR_ATTR AlarmClass_timeup(void)
{
  AlarmTimer.timeup();
}

bool Hardware::setupAlarmTimer(void)
{
  timer = timerBegin(ALARM_TIMER_ID, F_CPU <= 80000000 ? F_CPU / 1000000 : 80, true);
  timerAttachInterrupt(timer, AlarmClass_timeup, true);
  return true;
}

void Hardware::startAlarmTimer(int32_t interval)
{
  timerAlarmDisable(timer);
  timerWrite(timer, 0);
  timerAlarmWrite(timer, interval < 2 ? 2 : interval, false);
  timerAlarmEnable(timer);
}

size_t Hardware::disableAndSaveInterrupts(void)
{
  return portSET_INTERRUPT_MASK_FROM_ISR();
}

void Hardware::restoreInterrupts(size_t save)
{
  portCLEAR_INTERRUPT_MASK_FROM_ISR(save);
}

void Hardware::idle(void)
{
  portENABLE_INTERRUPTS();
}

#endif
