/*
  hardware.cpp - Hardware Library for RP2040

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include "../../config.h"

#ifdef __ESP8266__

#include "../../hardware.h"
#include "../../alarm.h"

#ifndef ALARM_TIMER_ID
#define ALARM_TIMER_ID 1  // TIMER1 is used exclusively with PWM
#endif

#define TIMER_US   (F_CPU / 1000000UL)
#define TIMER0_MAX (0xFFFFFFFFUL / TIMER_US)
#define TIMER1_MAX (0x007FFFFFUL / TIMER_US)

void ISR_ATTR AlarmClass_timeup(void)
{
  AlarmTimer.timeup();
}

void Hardware::setupAlarmTimer(void)
{
#if ALARM_TIMER_ID == 0
  timer0_isr_init();
  timer0_attachInterrupt(AlarmClass_timeup);
  timer0_write(ESP.getCycleCount());
#elif ALARM_TIMER_ID == 1
  timer1_isr_init();
  timer1_attachInterrupt(AlarmClass_timeup);
  timer1_enable(TIM_DIV1, TIM_EDGE, TIM_SINGLE);
#endif
}

bool Hardware::startAlarmTimer(int32_t interval)
{
#if ALARM_TIMER_ID == 0
  timer0_write((interval < 2 ? 2 : ((uint32_t)interval >= TIMER0_MAX ? TIMER0_MAX : interval)) * TIMER_US + ESP.getCycleCount());
#elif ALARM_TIMER_ID == 1
  timer1_write((interval < 2 ? 2 : ((uint32_t)interval >= TIMER1_MAX ? TIMER1_MAX : interval)) * TIMER_US);
#endif
  return true;
}

size_t Hardware::disableAndSaveInterrupts(void)
{
  return noInterrupts();
}

void Hardware::restoreInterrupts(size_t save)
{
  xt_wsr_ps(save);
}

void Hardware::idle(void)
{
  interrupts();
}

#endif
