/*
  hardware.cpp - Hardware Library for nRF52840

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can rdistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include "../../config.h"

#ifdef __NRF52840__

#include <nrf.h>
#include "../../hardware.h"
#include "../../alarm.h"

#ifndef ALARM_TIMER_ID
#define ALARM_TIMER_ID 4
#endif

#if ALARM_TIMER_ID == 0
  #define TIMER_REG  NRF_TIMER0
  #define TIMER_IRQn TIMER0_IRQn
#elif ALARM_TIMER_ID == 1
  #define TIMER_REG  NRF_TIMER1
  #define TIMER_IRQn TIMER1_IRQn
#elif ALARM_TIMER_ID == 2
  #define TIMER_REG  NRF_TIMER2
  #define TIMER_IRQn TIMER2_IRQn
#elif ALARM_TIMER_ID == 3
  #define TIMER_REG  NRF_TIMER3
  #define TIMER_IRQn TIMER3_IRQn
#elif ALARM_TIMER_ID == 4
  #define TIMER_REG  NRF_TIMER4
  #define TIMER_IRQn TIMER4_IRQn
#else
  #error "ALARM_TIMER_ID is invalid"
#endif

#define TIMER_PRIO APP_IRQ_PRIORITY_LOW

void AlarmClass_timeup(void)
{
  TIMER_REG->EVENTS_COMPARE[0] = 0;
  AlarmTimer.timeup();
}

bool Hardware::setupAlarmTimer(void)
{
  TIMER_REG->TASKS_STOP        = 1;
  TIMER_REG->TASKS_CLEAR       = 1;
  TIMER_REG->MODE              = TIMER_MODE_MODE_Timer;
  TIMER_REG->BITMODE           = TIMER_BITMODE_BITMODE_32Bit;
  TIMER_REG->PRESCALER         = 4; // 1MHz = 16MHz / (1 << 4)
  TIMER_REG->SHORTS            = 0;
  TIMER_REG->INTENCLR          = -1;
  TIMER_REG->INTENSET          = TIMER_INTENSET_COMPARE0_Set << TIMER_INTENSET_COMPARE0_Pos;
  TIMER_REG->EVENTS_COMPARE[0] = 0;
  TIMER_REG->CC[0]             = 0;
  NVIC_SetVector      (TIMER_IRQn, (uint32_t)AlarmClass_timeup);
  NVIC_SetPriority    (TIMER_IRQn, TIMER_PRIO);
  NVIC_ClearPendingIRQ(TIMER_IRQn);
  NVIC_EnableIRQ      (TIMER_IRQn);
  TIMER_REG->TASKS_START       = 1;
  return true;
}

void Hardware::startAlarmTimer(int32_t interval)
{
  TIMER_REG->TASKS_CAPTURE[1] = 1;
  TIMER_REG->CC[0] = TIMER_REG->CC[1] + (interval < 2 ? 2 : interval);
}

size_t Hardware::disableAndSaveInterrupts(void)
{
  size_t save = __get_PRIMASK();
  __disable_irq();
  return save;
}

void Hardware::restoreInterrupts(size_t save)
{
  __set_PRIMASK(save);
}

void Hardware::idle(void)
{
  __enable_irq();
  __WFI();
}

#endif
