/*
  hardware.cpp - Hardware Library for RP2040

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include "../../config.h"

#ifdef __RP2040__

#include <hardware/irq.h>
#include <hardware/structs/timer.h>
#include <hardware/timer.h>
#include "../../hardware.h"
#include "../../alarm.h"

#ifndef ALARM_TIMER_ID
#define ALARM_TIMER_ID 2
#endif

inline void AlarmClass_timeup(void)
{
  timer_hw->intr = 1 << ALARM_TIMER_ID;
  AlarmTimer.timeup();
}

void TimerCallback(void)
{
  AlarmClass_timeup();
}

bool Hardware::setupAlarmTimer(void)
{
  irq_set_exclusive_handler(ALARM_TIMER_ID + TIMER_IRQ_0, TimerCallback);
  timer_hw->intr = 1 << ALARM_TIMER_ID;
  hw_set_bits(&timer_hw->inte, 1 << ALARM_TIMER_ID);
  irq_clear(ALARM_TIMER_ID + TIMER_IRQ_0);
  irq_set_enabled(ALARM_TIMER_ID + TIMER_IRQ_0, true);
  return true;
}

void Hardware::startAlarmTimer(int32_t interval)
{
  timer_hw->armed = 1 << ALARM_TIMER_ID;
  timer_hw->intr  = 1 << ALARM_TIMER_ID;
  timer_hw->alarm[ALARM_TIMER_ID] = (interval < 2 ? 2 : interval) + time_us_32();
}

size_t Hardware::disableAndSaveInterrupts(void)
{
  size_t save = __get_PRIMASK();
  __disable_irq();
  return save;
}

void Hardware::restoreInterrupts(size_t save)
{
  __set_PRIMASK(save);
}

void Hardware::idle(void)
{
  __enable_irq();
  __WFI();
}

#endif
