/*
  hardware.cpp - Hardware Library for RP2040

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include "../../config.h"

#ifdef __STM32__

#include "HardwareTimer.h"
#include "../../hardware.h"
#include "../../alarm.h"

#ifndef ALARM_TIMER_ID
  #define ALARM_TIMER_ID 3
#endif

#undef TIMER
#if ALARM_TIMER_ID == 1
  #if defined(TIM1_BASE)
    #define TIMER TIM1
  #endif
#elif ALARM_TIMER_ID == 2
  #if defined(TIM2_BASE)
    #define TIMER TIM2
  #endif
#elif ALARM_TIMER_ID == 3
  #if defined(TIM3_BASE)
    #define TIMER TIM3
  #endif
#elif ALARM_TIMER_ID == 4
  #if defined(TIM4_BASE)
    #define TIMER TIM4
  #endif
#elif ALARM_TIMER_ID == 5
  #if defined(TIM5_BASE)
    #define TIMER TIM5
  #endif
#elif ALARM_TIMER_ID == 6
  #if defined(TIM6_BASE)
    #define TIMER TIM6
  #endif
#elif ALARM_TIMER_ID == 7
  #if defined(TIM7_BASE)
    #define TIMER TIM7
  #endif
#elif ALARM_TIMER_ID == 8
  #if defined(TIM8_BASE)
    #define TIMER TIM8
  #endif
#elif ALARM_TIMER_ID == 9
  #if defined(TIM9_BASE)
    #define TIMER TIM9
  #endif
#elif ALARM_TIMER_ID == 10
  #if defined(TIM10_BASE)
    #define TIMER TIM10
  #endif
#elif ALARM_TIMER_ID == 11
  #if defined(TIM11_BASE)
    #define TIMER TIM11
  #endif
#elif ALARM_TIMER_ID == 12
  #if defined(TIM12_BASE)
    #define TIMER TIM12
  #endif
#elif ALARM_TIMER_ID == 13
  #if defined(TIM13_BASE)
    #define TIMER TIM13
  #endif
#elif ALARM_TIMER_ID == 14
  #if defined(TIM14_BASE)
    #define TIMER TIM14
  #endif
#elif ALARM_TIMER_ID == 15
  #if defined(TIM15_BASE)
    #define TIMER TIM15
  #endif
#elif ALARM_TIMER_ID == 16
  #if defined(TIM16_BASE)
    #define TIMER TIM16
  #endif
#elif ALARM_TIMER_ID == 17
  #if defined(TIM17_BASE)
    #define TIMER TIM17
  #endif
#elif ALARM_TIMER_ID == 18
  #if defined(TIM18_BASE)
    #define TIMER TIM18
  #endif
#elif ALARM_TIMER_ID == 19
  #if defined(TIM19_BASE)
    #define TIMER TIM19
  #endif
#elif ALARM_TIMER_ID == 20
  #if defined(TIM20_BASE)
    #define TIMER TIM20
  #endif
#elif ALARM_TIMER_ID == 21
  #if defined(TIM21_BASE)
    #define TIMER TIM21
  #endif
#elif ALARM_TIMER_ID == 22
  #if defined(TIM22_BASE)
    #define TIMER TIM22
  #endif
#endif

#define MAX_RELOAD ((1 << 16) - 1)
#define CHANNEL    1

static HardwareTimer timer(TIMER);

void AlarmClass_timeup(void)
{
#if defined(TIMER)
  TIMER->SR = 0;
#endif
  AlarmTimer.timeup();
}

bool Hardware::setupAlarmTimer(void)
{
#if defined(TIMER)
  timer.setPrescaleFactor(timer.getTimerClkFreq() / 1000000);
  timer.attachInterrupt(CHANNEL, AlarmClass_timeup);
  timer.setMode(CHANNEL, TIMER_DISABLED);
  HAL_TIM_OC_Start(timer.getHandle(), CHANNEL);
  return true;
#endif
  return false;
}

void Hardware::startAlarmTimer(int32_t interval)
{
#if defined(TIMER)
  timer.setCaptureCompare(CHANNEL, (interval < 2 ? 2 : ((uint32_t)interval >= MAX_RELOAD ? MAX_RELOAD : interval)) + timer.getCount());
#endif
}

size_t Hardware::disableAndSaveInterrupts(void)
{
  size_t save = __get_PRIMASK();
  __disable_irq();
  return save;
}

void Hardware::restoreInterrupts(size_t save)
{
  __set_PRIMASK(save);
}

void Hardware::idle(void)
{
  __enable_irq();
  __WFI();
}

#endif
